/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: George F. Riley<riley@ece.gatech.edu>
 * Modifications Lukasz Kiedrowski <kiedrowski@itti.com.pl>
 * Based on ANIMATION_INTERFACE__H (AnimationInterface.cc)
 */

// Interface between ns3 and the network animator

#include <stdio.h>
#include <sstream>

#include "ns3/net-anim-config.h"

// Socket related includes
#if defined(HAVE_SYS_SOCKET_H) && defined(HAVE_NETINET_IN_H)
#include <sys/socket.h>
#include <netinet/in.h>
#else
#include <fcntl.h>
#endif

// ns3 includes
#include "ns3/channel.h"
#include "ns3/config.h"
#include "ns3/node.h"
#include "ns3/canvas-location.h"
#include "ns3/packet.h"
#include "ns3/simulator.h"

#include "dtn-tool.h"  //LKIE itti 300910
#include "ns3/vector.h"  //LKIE itti 300910
#include <fstream>       //LKIE itti 081010

using namespace std;

NS_LOG_COMPONENT_DEFINE ("DTNTool");

namespace ns3 {

DTNTool::DTNTool ()
  : m_fHandle (STDOUT_FILENO), m_model (0), m_port(0) ,m_circle_radius(0) //300910 added
{
}

DTNTool::~DTNTool ()
{
}

bool DTNTool::SetOutputFile (const std::string& fn)
{
  m_port = 0;
  FILE* f = fopen (fn.c_str (), "w");
  if (!f)
    {
      return false; // Can't open
    }
  m_fHandle = fileno (f); // Set the file handle
  return true;
}

bool DTNTool::SetServerPort (uint16_t port)
{
  m_port = port; //LKIE itti 300910
#if defined(HAVE_SYS_SOCKET_H) && defined(HAVE_NETINET_IN_H)
  int s = socket (AF_INET, SOCK_STREAM, 0);
  struct sockaddr_in addr;
  addr.sin_family = AF_INET;
  addr.sin_port = htons (port);
  addr.sin_addr.s_addr = htonl (INADDR_ANY);
  if (bind (s, (struct sockaddr*)&addr, sizeof (addr)) < 0)
    {
      NS_LOG_WARN ("Can't bind to port " << port << ", exiting.");
      return false;
    }
  listen (s, 1);
  NS_LOG_INFO ("Waiting for animator connection");
  // Now wait for the animator to connect in
  m_fHandle = accept (s, 0, 0);
  NS_LOG_INFO ("Got animator connection from remote");
  // set the linger socket option
  int t = 1;
  setsockopt (s, SOL_SOCKET, SO_LINGER, &t, sizeof(t));
  return true;
#endif
  return false; // never reached unless the above is disabled
                // which is done to support a platform like MinGW
}

void DTNTool::SetRadius (double radius) //LKIE itti 300910
{
   m_circle_radius = radius;
}

MapNam
DTNTool::StartAnimationWifi (uint32_t vbox_no, uint32_t vpn_no, uint32_t lxc_no, uint32_t lxc_mob_no) //LKIE itti 300910
{
  std::ostringstream ossn;
  Time now = Simulator::Now ();
  //set circle radius
  ossn.str(""); 
  if ( m_port > 0)
  {
    ossn << "0.0 R " << m_circle_radius << endl;
  }
  else 
  {
    ossn << now.GetSeconds() << " R " << m_circle_radius << endl;
  }
  //WriteNSimple (ossn.str ().c_str (), ossn.str ().length ());
  WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());

  //update nodes
  Vector3D vi, vj;
  for (NodeList::Iterator i = NodeList::Begin (); i != NodeList::End (); ++i)
  { 
    Ptr<Node> n = *i;
    vi=(n->GetObject<MobilityModel>())->GetPosition();
    uint32_t nId = n->GetId ();
    ossn.str("");
    char x;
    if ((n->GetId ()) < vbox_no) x='r';
    else if ( ((nId) >= vbox_no) && ((nId) < (vbox_no + vpn_no)) ) x='g';
    else if ( ((nId) >= (vbox_no + vpn_no)) && ((nId) < (vbox_no + vpn_no + (lxc_no - lxc_mob_no) )) ) x='b';
    else x='m';
   
    if ( m_port > 0)
    {
      ossn << "0.0 N " << nId << " " << vi.x << " " << vi.y << " " << x <<endl;
    }
    else 
    {
      ossn << now.GetSeconds() << " N " << nId << " " << vi.x << " " << vi.y << " " << x <<endl;
    }
    //anim.WriteNSimple (ossn.str ().c_str (), ossn.str ().length ());
    WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());
    
  }
  //update links
  uint32_t ni, nj;
  Ptr<MobilityModel> vi_m, vj_m;

 for (NodeList::Iterator i = NodeList::Begin (); i != NodeList::End (); ++i)
    {
      Ptr<Node> n_i = *i;
      ///--------assumed that all nodes are wireless
      //vi=(n_i->GetObject<MobilityModel>())->GetPosition();
      vi_m=(n_i->GetObject<MobilityModel>());
      ni=n_i->GetId ();
      for (NodeList::Iterator j = i+1; j != NodeList::End (); ++j)
      {
      	Ptr<Node> n_j = *j;
        //vj=(n_j->GetObject<MobilityModel>())->GetPosition();
        vj_m=(n_j->GetObject<MobilityModel>());
        nj=n_j->GetId ();
        //if ( ((vi.x-vj.x)*(vi.x-vj.x) + (vi.y-vj.y)*(vi.y-vj.y)) <= (m_circle_radius*m_circle_radius) )
        if ( (vi_m->GetDistanceFrom (vj_m)) <= m_circle_radius )
        {
       		ossn.str(""); 
                if (m_port > 0) ossn << "0.0 L "  << ni << " " << nj << endl;
                else ossn << now.GetSeconds() << " L "  << ni << " " << nj << endl;
                    
                //anim.WriteNSimple (ossn.str ().c_str (), ossn.str ().length ());
                WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());
         }
      }
    }
  ossn.str("");
  ossn << "/NodeList/" << "*" << "/$ns3::MobilityModel/CourseChange";
  Config::Connect (ossn.str (), MakeCallback (&DTNTool::CourseChanged, this));
  
  return MapNames(vbox_no, vpn_no, lxc_no);
}

MapNam DTNTool::MapNames (uint32_t vbox_no, uint32_t vpn_no, uint32_t lxc_no)
{
  MapNam id_to_real; //heve made it global 190111
  uint32_t nodes_total = lxc_no+vpn_no+vbox_no;

  //---map nodes	    ns3_id       	to        real ip address ending according naming rules
  /*          (0;vbox_no -1)        		->	  (1;vbox_no)
              (0;vbox_no + vpn_no -1)		->	  (201;200 + vpn_no)
              (0;vbox_no + vpn_no + lxc_no -1)  ->	  (101;100 + lxc_no)
  */
  for (uint32_t ns3_id=0; ns3_id<nodes_total; ns3_id++)
  { 
    if (ns3_id < vbox_no) 		 id_to_real[ns3_id] = ns3_id + 1;
    else if (ns3_id < (vbox_no + vpn_no))id_to_real[ns3_id] = ns3_id - vbox_no + 200 +1;
    else 				 id_to_real[ns3_id] = ns3_id - vbox_no - vpn_no + 100 +1;
    //NS_LOG_INFO("ns3_id [" << ns3_id << "]= "<< id_to_real[ns3_id]);
  }

  m_map_names = id_to_real;
  std::ostringstream ossn;
  Time now = Simulator::Now ();
  //update nodes names
  //for ( uint32_t ns3_id = 0; ns3_id < id_to_real.size (); ns3_id++)
  for ( MapNam::iterator ns3_id = id_to_real.begin (); ns3_id != id_to_real.end (); ++ns3_id)
  { 
    ossn.str("");  
    if ( m_port > 0)
    {
      //ossn << "0.0 M " << ns3_id << " " << id_to_real[ns3_id] << endl;
        ossn << "0.0 M " << ns3_id->first << " " <<  ns3_id->second << endl;
    }
    else 
    {
      //ossn << now.GetSeconds() << " M " << ns3_id << " " << id_to_real[ns3_id] << endl;
      ossn << now.GetSeconds() << " M " << ns3_id->first << " " <<  ns3_id->second << endl;
    }
    //NS_LOG_INFO("ns3_id [" << ns3_id << "]= "<< id_to_real[ns3_id]);
    //NS_LOG_INFO("ns3_id [" << ns3_id->first << "]= "<< ns3_id->second);
    //cout << "ns3_id [" << ns3_id->first << "]= "<< ns3_id->second << endl ;
    WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());
  }
  return id_to_real;
}
void
DTNTool::CourseChangedVelocity (const NodeContainer* nc)//(Ptr<const NodeContainer> nc)
{
  std::ostringstream oss;
  Ptr<Node> n;
  for (uint32_t i=0; i<nc->GetN(); i++) //only odd nodes
  {
  n = nc->Get (i);
  oss.str("");
  oss << "/NodeList/" << n->GetId () << "/$ns3::MobilityModel/CourseChange";
  std::string context = oss.str ();
  DTNTool::CourseChanged (context, n->GetObject<MobilityModel>());
  }
}

void
DTNTool::CourseChanged (std::string context, Ptr<const MobilityModel> model)
{
  Time now = Simulator::Now ();
 //if (now.GetSeconds() > 0){
  Vector position = model->GetPosition ();
  //std::vector<std::string> splitPath = PathSplit (context);
  std::vector<std::string> splitPath = PathSplit (context);
  uint32_t nodeIndex = atoi (splitPath[1].c_str ());
  //cout<<"node id : "<<nodeIndex << endl; 
  std::ostringstream ossn;
  
  //update nodes
  ossn.str("");
  if (m_port > 0)
  {
   ossn << "0.0 N " << nodeIndex << " " << position.x << " " << position.y << endl; //" g" <<endl;
  }
  else 
  {
   ossn << now.GetSeconds() << " N " << nodeIndex << " " << position.x << " " << position.y << endl; //" g" <<endl;
  }
  //anim.WriteNSimple (ossn.str ().c_str (), ossn.str ().length ());
  WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());
  //del previous links
  ossn.str(""); 

  if (m_port > 0) ossn << "0.0 D " << nodeIndex << endl;
  else ossn << now.GetSeconds() << " D " << nodeIndex << endl;

  //anim.WriteNSimple (ossn.str ().c_str (), ossn.str ().length ());
  WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());
  //update links
  uint32_t nj;
  Ptr<MobilityModel> vj_m;
  for (NodeList::Iterator j = NodeList::Begin (); j != NodeList::End (); ++j)
  {
    Ptr<Node> n_j = *j;
    nj=n_j->GetId ();
    if (nodeIndex != nj){
    vj_m=(n_j->GetObject<MobilityModel>());
    if ( (model->GetDistanceFrom (vj_m)) <= m_circle_radius )
    {
     ossn.str(""); 
     if (m_port > 0)
     {
      ossn << "0.0 L "  << nodeIndex << " " << nj << endl;
     }
     else 
     {
      ossn << now.GetSeconds() << " L "  << nodeIndex << " " << nj << endl;
     }
     //anim.WriteNSimple (ossn.str ().c_str (), ossn.str ().length ());
     WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());
    }
   }
  }
//}
}

void DTNTool::StartAnimation ()
{
  // Dump the topology
  for (NodeList::Iterator i = NodeList::Begin (); i != NodeList::End (); ++i)
    {
      Ptr<Node> n = *i;
      Ptr<CanvasLocation> loc = n->GetObject<CanvasLocation> ();
      if (loc)
        {
          // Location exists, dump it
          Vector v = loc->GetLocation ();
          ostringstream oss;
          oss << "0.0 N " << n->GetId () 
               << " " << v.x << " " << v.y << endl;
          WriteN (m_fHandle, oss.str ().c_str (), oss.str ().length ());
        }
    }
  // Now dump the p2p links
  for (NodeList::Iterator i = NodeList::Begin (); i != NodeList::End (); ++i)
    {
      Ptr<Node> n = *i;
      uint32_t n1Id = n->GetId ();
      uint32_t nDev = n->GetNDevices ();  // Number of devices
      for (uint32_t i = 0; i < nDev; ++i)
        {
          Ptr<NetDevice> dev = n->GetDevice (i);
          Ptr<Channel>   ch = dev->GetChannel ();
          if (!ch) 
            {
              continue; // No channel, can't be p2p device
            }
          string channelType = ch->GetInstanceTypeId ().GetName ();
          if (channelType == string ("ns3::PointToPointChannel"))
            { // Since these are duplex links, we only need to dump
              // if srcid < dstid
              uint32_t nChDev = ch->GetNDevices ();
              for (uint32_t j = 0; j < nChDev; ++j)
                {
                  Ptr<NetDevice> chDev = ch->GetDevice (j);
                  uint32_t n2Id = chDev->GetNode ()->GetId ();
                  if (n1Id < n2Id)
                    { // ouptut the p2p link
                      ostringstream oss;
                      oss << "0.0 L "  << n1Id << " " << n2Id << endl;
                      WriteN (m_fHandle, oss.str ().c_str (),
                             oss.str ().length ());
                    }
                }
            }
          else
            {//LKIE itti 220910
              //NS_FATAL_ERROR ("Net animation currently dont support that kind of links.");
              NS_LOG_INFO ("No pointToPoint links; use StartAnimationWifi (uint32_t, uint32_t, uint32_t)");
            }
        }
    }
  
  // Connect the callback for packet tx events
  Config::Connect ("/ChannelList/*/TxRxPointToPoint",
                   MakeCallback (&DTNTool::DevTxTrace, this));
}

void DTNTool::StopAnimation ()
{
  if (m_fHandle > 0) 
    {
      close (m_fHandle);
    }
}
// From n4c

Vector3D*
DTNTool::ReadPositions (const std::string filename)
{

  ifstream topgen_tmp;
  //topgen_tmp.open (GetFileName ().c_str ());
  topgen_tmp.open (filename.c_str ());
  double x=0.0;
  double y=0.0;
  double z=0.0;

  string lin;
  int j;
  for (j = 0; !topgen_tmp.eof (); j++)
  {
    getline (topgen_tmp,lin);
  }
  int sp_no = int(j/3);
  Vector3D space_point[3*sp_no]; //memory alloc !!! 3xdouble var !!!
  
  topgen_tmp.close ();
  //topgen_tmp.open (GetFileName ().c_str ());

  istringstream lineBuffer;
  string line;
  ifstream topgen;
  //topgen.open (GetFileName ().c_str ());
  topgen.open (filename.c_str ());  

  string node;
  string command;
  string dim;
  

    int i = 0;
    while (i<sp_no && !topgen.eof ())
    {
      for (int k = 1; k <= 3; k++)
      {
        line.clear ();
        lineBuffer.clear ();
        node.clear ();
        command.clear ();
        dim.clear ();
        double pos=0;
 
        getline (topgen,line);
        lineBuffer.str (line);

        lineBuffer >> node >> command >> dim >> pos; 

        if ( (!node.empty ()) && (!dim.empty ()) && (!command.empty ()) )
        {
         if (k == 3)
         {
                z = pos;
                space_point[i] = Vector3D(x, y, z);
                //cout<<i<<"       "<<space_point[i]<<endl;
                i++;
         } 
         else if (k == 2)
         {
                y = pos;
         }
         else 
         {
                x = pos;
         }
        }
       else 
       {
        cout<<"empty line?"<<endl;
        k=k-1;
       }

      }
     
    }

  cout<<sp_no << " spacepoints created based on ns2 topology "<<endl;
  topgen.close ();
  //const Vector3D* sp = &space_point[0];
  Vector3D* sp = space_point;

  return sp;
}

MeetEvents
DTNTool::ReadMovement(const std::string filename)//LKIE itti 170111
{

  MeetEvents meet_events;

  istringstream lineBuffer;
  string line;
  ifstream topgen;
  //topgen.open (GetFileName ().c_str ());
  topgen.open (filename.c_str ());  

  getline (topgen,line);
  lineBuffer.str (line);
  int node_number;
  lineBuffer >> node_number;   
  cout<<"Read_movement () - > File include MEETING LIST for "<< node_number <<" nodes"<<endl;

  movement m_e;
  int i = 0;
  while (!topgen.eof ())
  {

        line.clear ();
        lineBuffer.clear ();
 
        getline (topgen,line);
        lineBuffer.str (line);
        lineBuffer >> m_e.n1 >> m_e.n2 >> m_e.t1 >> m_e.t2;   
        //meet_events[i] = m_e; // until initialized it is wrong - remember!!!!!
        meet_events.push_back(m_e);
        //meet_events[i] = struct movement(n1, n2, t1 , t2);
        i++;
  }

  cout<<"Read_movement () - > " << i <<" node meetings created based on DTN characteristic movement generator"<<endl;
  topgen.close ();

  return meet_events;
}

void
DTNTool::InitializeInteractiveTerminal (const std::string user, const std::string path_interactive)
{
//---Interactive console
  ostringstream ossn;
  string command = "\"gnome-terminal -e \"";
  ossn.str("");
  ossn << "su " << user.c_str () <<" -c " << command.c_str () << path_interactive.c_str () << "it\"\""; //Fedora
  string command_ = ossn.str();
  system(command_.c_str());

  ossn.str("");
  ossn << path_interactive.c_str () << "command";
  m_interactive_file = ossn.str();

  ofstream fout(m_interactive_file.c_str () );
  fout << -1 << " " << 0 << " "<< 0;  // initialization
  fout.close();
}

void
DTNTool::ChangeNodeAttr ()
{
  //uint32_t node_id=0;
  char action;
  int node_id =0;
  //int ip_id =0;
  double x=0.0;
  double y=0.0;

  std::string filename = m_interactive_file;

  istringstream lineBuffer;
  string line;
  ifstream topgen;
  topgen.open (filename.c_str ());  

  getline (topgen,line);
  lineBuffer.str (line);
  lineBuffer >> action;

  //cout<<node_id << " " << x << " " << y << " readed"<<endl;  
  //static Ptr< Node > 	GetNode (uint32_t n)

  switch (action) 
    {
    case 'n' : // change pos according to ns3 Node id
      lineBuffer >> node_id >> x >> y;
      topgen.close ();
      if (( node_id = FindKeyByVal(uint32_t(node_id)) ) >=0)
      {
         Ptr<Node> n_ptr = NodeList::GetNode (uint32_t(node_id));
         Vector3D v(x, y, 0);
         (n_ptr->GetObject<MobilityModel>())->SetPosition(v);
       }else cout << "Wrong node number" << endl;
      break;
    case 'l' : // on/off label
      lineBuffer >> node_id;
      topgen.close ();
      if ( (node_id = FindKeyByVal(uint32_t(node_id)) ) >=0)
      {
        Time now = Simulator::Now ();
        std::ostringstream ossn;
        ossn.str("");
        if (m_port > 0)
        {
          ossn << "0.0 " << action << " " << node_id << endl;
        }
        else 
        {
          ossn << now.GetSeconds() << " " << action << " " << node_id << endl;
        }
        WriteN (m_fHandle, ossn.str ().c_str (), ossn.str ().length ());
      }else cout << "Wrong node number" << endl;
      break;
    default:
      break;
    }
    ofstream fout( filename.c_str () );
    fout << -1;
    fout.close();
}

uint32_t DTNTool::FindKeyByVal(uint32_t value)
{
  //std::map<int, std::string>::const_iterator it;
  uint32_t key = -1;

  for (MapNam::iterator it = m_map_names.begin(); it != m_map_names.end(); ++it)
  {
    if (it->second == value)
    {
    key = it->first;
    break;
    }
  }
  return key;
}

std::vector<std::string>
DTNTool::PathSplit (std::string str)
{
  std::vector<std::string> results;
  size_t cutAt;
  while ((cutAt = str.find_first_of('/')) != str.npos)
    {
      if(cutAt > 0)
        {
          results.push_back(str.substr(0,cutAt));
        }
      str = str.substr(cutAt+1);
    }
  if (str.length() > 0)
    {
      results.push_back(str);
    }
  return results;
}

// Private methods
int DTNTool::WriteN (int h, const char* data, uint32_t count)
{ // Write count bytes to h from data
  uint32_t    nLeft   = count;
  const char* p       = data;
  uint32_t    written = 0;

  while (nLeft)
    {
      int n = write (h, p, nLeft);
      if (n <= 0) 
        {
          return written;
        }
      written += n;
      nLeft -= n;
      p += n;
    }
  return written;
}

//LKIE itti 220910
int DTNTool::WriteNSimple (const char* data, uint32_t count)
{ 
  uint32_t    written = 0;
  written = WriteN (m_fHandle, data, count);
  return written;
}
  
void DTNTool::DevTxTrace (std::string context, Ptr<const Packet> p,
                                     Ptr<NetDevice> tx, Ptr<NetDevice> rx,
                                     Time txTime, Time rxTime)
{
  Time now = Simulator::Now ();
  ostringstream oss;
  oss << now.GetSeconds() << " P "
      << tx->GetNode ()->GetId () << " "
      << rx->GetNode ()->GetId () << " "
      << (now + txTime).GetSeconds () << " " // last bit tx time
      << (now + rxTime - txTime).GetSeconds() << " "  // first bit rx time
      << (now + rxTime).GetSeconds () << endl;         // last bit rx time
  WriteN (m_fHandle, oss.str ().c_str (), oss.str ().length ());
}

} // namespace ns3
